"""GearClient
**********

Client for the APIs provided by Homegear.
"""

import future  # noqa

from .methods import SystemMethodsCollection, DeviceMethodsCollection
from .protocols import XMLRPC, initialise_protocol


class GearClient(object):
    """Client for the APIs provided by Homegear.

    Usage:

        >>> gc = GearClient('localhost', 1234)
        >>> gc.device.list_devices()

    The default communication protocol is ``XML RPC``. Additionally, ``JSON RPC`` is supported:

        >>> from gearthonic import JSONRPC
        >>> GearClient('localhost', 1234, protocol=JSONRPC)

    Any protocol can accept additional parameters. Provide them while initialising the client:

        >>> GearClient('localhost', 1234, protocol=JSONRPC, secure=False, username='ham')

    For a full list of supported parameters, have a look at each protocol class or see the documentation for the
    protocols. :class:
    """

    def __init__(self, host, port, protocol=XMLRPC, **kwargs):
        self.port = port
        self.host = host
        self.protocol_kwargs = kwargs
        self.protocol = self.__initialise_protocol(protocol)

        self.__system = SystemMethodsCollection(self)
        self.__device = DeviceMethodsCollection(self)

    @property
    def system(self):
        """Smart access to all system related API methods."""
        return self.__system

    @property
    def device(self):
        """Smart access to all device related API methods."""
        return self.__device

    def __initialise_protocol(self, protocol):
        """Initialise the given protocol.

        :param protocol: protocol to initialise
        :return: the protocol instance
        :rtype: gearthonic.protocols._ProtocolInterface
        """
        return initialise_protocol(protocol, self.host, self.port, **self.protocol_kwargs)

    def __getattr__(self, item):
        """Allow to call any method directly through the GearClient.

        :param item: the method name to call
        :type item: str
        :return: function which calls the given method
        :rtype: function
        """
        def wrapper(*args, **kwargs):
            return self.call(item, *args, **kwargs)
        return wrapper

    def call(self, method_name, *args, **kwargs):
        """Call the given method using the instantiated protocol.

        :param method_name: method to be called
        :param args: arguments passed through
        :param kwargs: keyword arguments passed through
        :return: return value of the call
        """
        return self.protocol.call(method_name, *args, **kwargs)



"""Includes all functions provided by the XML RPC API, split into logical entities."""


class BaseCollection(object):
    """Provide the interface to call XML RPC methods."""

    def __init__(self, caller):
        """
        :param caller: interface to call XML RPC methods
        """
        self._caller = caller

    def call(self, method_name, *args, **kwargs):
        """Make a call to the XML RPC API.

        :param method_name: name of the method to call
        :type method_name: str
        :param args: arguments to pass to the method
        :param kwargs: keyword arguments to pass to the method
        """
        return self._caller.call(method_name, *args, **kwargs)


class SystemMethodsCollection(BaseCollection):
    """
    All XML RPC server provide a set of standard methods.
    """
    def get_capabilities(self):
        """Lists server's XML RPC capabilities.

        Example output::

            {
                'xmlrpc': {'specUrl': 'http://example.com', 'specVersion': 1}
                'faults_interop': {'specUrl': 'http://example2.com', 'specVersion': 101}
                'introspection': {'specUrl': 'http://example3.com', 'specVersion': 42}
            }

        :return: A dict containing all information
        :rtype: dict
        """
        return self.call('system.getCapabilities')

    def list_methods(self):
        """Lists servers's XML RPC methods.

        :return: A list of available methods
        :rtype: list
        """
        return self.call('system.listMethods')

    def method_help(self, method_name):
        """Returns the description of a method.

        :param method_name: The name of the method
        :type method_name: str
        :return: The description of the method
        :rtype: str
        """
        return self.call('system.methodHelp', method_name)

    def method_signature(self, method_name):
        """Returns the signature of a method.

        :param method_name: The name of the method
        :type method_name: str
        :return: The signature of the method
        :rtype: list
        """
        return self.call('system.methodSignature', method_name)

    def multicall(self, methods):
        """Call multiple  methods at once.

        Example list of ``methods``::

            [
                {'methodName': 'getValue', 'params': [13, 4, 'TEMPERATURE']},
                {'methodName': 'getValue', 'params': [3, 3, 'HUMIDITY']}
            ]

        Return value of the multicall::

            [22.0, 58]

        :param methods: A list of methods and their parameters
        :type methods: list
        :return: A list of method responses.
        :rtype: list
        """
        return self.call('system.multicall')


class DeviceMethodsCollection(BaseCollection):
    """All device related methods."""

    def activate_link_paramset(self, peer_id, channel, remote_peer_id, remote_channel, long_pressed=False):
        """
        :return:
        """
        return self.call('activateLinkParamset', peer_id, channel, remote_peer_id, remote_channel, long_pressed)

    def list_devices(self):
        """Return a list of devices.

        :return: List of devices
        :rtype: list
        """
        return self.call('listDevices')

    def get_value(self, peer_id, channel, key, request_from_device=False, asynchronous=False):
        """Return the value of the device, specified by channel and key (parameterName).

        Per default the value is read from the local cache of Homegear. If the value should be read from the device, use
        ``request_from_device``. If the value should be read from the device, this can be done asynchronously. The
        method returns immediately and doesn't wait for the current value. The value will be sent as an event as soon as
        it's returned by the device.

        Error codes:

        * Returns ``-2`` when the device or channel is unknown
        * Returns ``-5`` when the key (parameter) is unknown

        :param peer_id: ID of the device
        :type peer_id: int
        :param channel: Channel of the device to get the value for
        :type channel: int
        :param key: Name of the parameter to get the value for
        :type key: str
        :param request_from_device: If true value is read from the device
        :type request_from_device: bool
        :param asynchronous: If true value is read asynchronously
        :type asynchronous: bool
        :return: The value of the parameter or error code
        :rtype: unknown
        """
        return self.call('getValue', peer_id, channel, key, request_from_device, asynchronous)

    def set_value(self, peer_id, channel, key, value):
        """Set the value of the device, specified by channel and key (parameterName).

        :param peer_id: ID of the device
        :type peer_id: int
        :param channel: Channel of the device to set the value for
        :type channel: int
        :param key: Name of the parameter to get the value for
        :type key: str
        :param value: The value to set
        :type value: unknown
        :return: * ``None`` on success
                 * ``-2`` when the device or channel is unknown
                 * ``-5`` when the key (parameter) is unknown
                 * ``-100`` when the device did not respond
                 * ``-101`` when the device returns an error
        """
        return self.call('setValue', peer_id, channel, key, value)


class ServerMethodsCollection(BaseCollection):
    """All client server related methods."""

    def register_server(self, uri, name):
        """Register a server for callbacks.

        :param uri: URI of the server
        :param name: server name
        :return:
        """
        return self.call('init', uri, name)



"""Communication protocols
***********************

These classes are used to communicate with the different APIs provided by the Homegear server.
"""
from _ssl import CERT_REQUIRED
from ssl import SSLContext, CERT_NONE, PROTOCOL_SSLv23

from xmlrpc.client import ServerProxy

from jsonrpcclient import Request
from jsonrpcclient.http_server import HTTPServer

from .exceptions import ProtocolUnknown, ConfigurationError

XMLRPC = 0
JSONRPC = 1
MQTT = 2


class _ProtocolInterface(object):
    """Interface for all protocols."""

    def call(self, method_name, *args, **kwargs):
        raise NotImplementedError("You can't use the {} directly.".format(self.__class__.__name__))


class XmlRpcProtocol(_ProtocolInterface):
    """Communicate with Homegear via XML RPC.

        >>> xp = XmlRpcProtocol('host.example.com', 2003)
        >>> xp.call('listDevices')
        [...]

    Set ``secure=False`` to use ``http`` instead off ``https``. Set ``verify=False`` to skip the verification of the
    SSL cert.

    Provide credentials via ``username`` and ``password`` if the Homegear server is secured by basic auth. It's not
    possible to use authentication with an insecure connection (http)!
    """

    def __init__(self, host, port, secure=True, verify=True, username=None, password=None):
        self.secure = secure
        self.protocol = 'https' if self.secure else 'http'
        self.port = port
        self.host = host

        # Authentication required?
        auth = ''
        if username and password:
            if not self.secure:
                raise ConfigurationError("You can't use authentication with an insecure (http) connection.")
            auth = '{0}:{1}@'.format(username, password)
        self.authentication = bool(auth)

        self.uri = u'{0}://{1}{2}:{3}'.format(self.protocol, auth, self.host, self.port)

        # Create the XML RPC ServerProxy
        context = None
        if secure:
            context = SSLContext(PROTOCOL_SSLv23)
            context.verify_mode = CERT_REQUIRED if verify else CERT_NONE
        self.proxy = ServerProxy(self.uri, context=context)

    def call(self, method_name, *args, **kwargs):
        """Call the given method using the ServerProxy.

        :param method_name: Method to be called
        :param args: Arguments passed through
        :param kwargs: Keyword arguments passed through
        :return: Return value of the XML RPC method
        """
        return getattr(self.proxy, method_name)(*args, **kwargs)


class JsonRpcProtocol(_ProtocolInterface):
    """Communicate with Homegear via JSON RPC.

        >>> jp = JsonRpcProtocol('host.example.com', 2003)
        >>> jp.call('listDevices')
        [...]

    Set ``secure=False`` to use ``http`` instead off ``https``. Set ``verify=False`` to skip the verification of the
    SSL cert.

    Provide credentials via ``username`` and ``password`` if the Homegear server is secured by basic auth. It's not
    possible to use authentication with an insecure connection (http)!
    """

    def __init__(self, host, port, secure=True, verify=True, username=None, password=None):
        self.secure = secure
        self.protocol = 'https' if self.secure else 'http'
        self.host = host
        self.port = port
        self.uri = u'{0}://{1}:{2}'.format(self.protocol, self.host, self.port)
        self.server = HTTPServer(self.uri)
        if not verify:
            self.server.session.verify = False
        if username and password:
            if not self.secure:
                raise ConfigurationError("You can't use authentication with an insecure (http) connection.")
            self.server.session.auth = (username, password)
        self.authentication = bool(username and password and self.secure)

    def call(self, method_name, *args, **kwargs):
        """Call the given method using the HTTPServer.

        :param method_name: Method to be called
        :param args: Arguments passed through
        :param kwargs: Keyword arguments passed through
        :return: Return value of the XML RPC method
        """
        return self.server.send(Request(method_name, *args, **kwargs))


class _MqttProtocol(_ProtocolInterface):
    """MQTT protocol to communicate with Homegear via a MQTT message broker.

    Not yet finished."""


_PROTOCOL_MAPPING = {
    XMLRPC: XmlRpcProtocol,
    JSONRPC: JsonRpcProtocol,
    MQTT: _MqttProtocol
}


def initialise_protocol(protocol, host, port, **kwargs):
    """Factory method to initialise a specific protocol.

    :param protocol: ID of the protocol to initialise
    :type protocol: int
    :param host: host of the server
    :type host: str
    :param port: port of the server
    :type port: int
    :param kwargs: will be used to initialise the protocol
    :rtype: _ProtocolInterface
    """
    try:
        klass = _PROTOCOL_MAPPING[protocol]
    except KeyError:
        raise ProtocolUnknown("Protocol with ID {} unknown.".format(protocol))
    return klass(host, port, **kwargs)



from .client import GearClient  # noqa
from .protocols import XMLRPC, JSONRPC, JsonRpcProtocol, XmlRpcProtocol  # noqa

__author__ = 'Timo Steidle'
__email__ = 'mumpitz@wumpitz.de'
__version__ = '0.2.0'



"""Tests for `gearthonic` module."""

from mock import Mock
from gearthonic import client


class TestGearClient(object):
    """Tests for the base GearClient"""

    def test_getttr_function(self, gear_client):
        """Test that __getattr__ returns a callable"""
        fnc = gear_client.ham
        assert callable(fnc)

    def test_magic_call(self, gear_client):
        """Test that any call gets redirected to the ``call`` method of the GearClient."""
        gear_client.call = Mock()
        gear_client.list_all(1, ham='eggs')

        assert gear_client.call.called
        gear_client.call.assert_called_once_with('list_all', 1, ham='eggs')

    def test_call_method(self, gear_client):
        """Test that the call method redirects any call to the XML RPC ServerProxy."""
        gear_client.protocol = Mock()
        gear_client.call('list_all', 1, ham='eggs')

        assert gear_client.protocol.call.called
        gear_client.protocol.call.assert_called_once_with('list_all', 1, ham='eggs')

    def test_protocol_initialised(self):
        """Test that the factory function for creating a protocol will be called."""
        client.initialise_protocol = Mock(return_value=123)

        gc = client.GearClient('localhost', 1234, protocol=55, ham='eggs')
        assert gc.protocol == 123
        client.initialise_protocol.assert_called_once_with(55, 'localhost', 1234, ham='eggs')



#!/usr/bin/env python

import os
import sys

try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup


if sys.argv[-1] == 'publish':
    os.system('python setup.py sdist upload')
    sys.exit()

readme = open('README.rst').read()
doclink = """
Documentation
-------------

The full documentation can be found at http://gearthonic.rtfd.org."""
changelog = open('CHANGELOG.rst').read().replace('.. :changelog:', '')

setup(
    name='gearthonic',
    version='0.2.0',
    description='A simple client to the XML RPC API of Homegear.',
    long_description=readme + '\n\n' + doclink + '\n\n' + changelog,
    author='Timo Steidle',
    author_email='mumpitz@wumpitz.de',
    url='https://gitlab.com/wumpitz/gearthonic',
    packages=[
        'gearthonic',
    ],
    package_dir={'gearthonic': 'gearthonic'},
    include_package_data=True,
    install_requires=[
        'future',
        'jsonrpcclient>=2.1.1,<3',
        'requests>=2.0.0,<3'
    ],
    license='MIT',
    zip_safe=False,
    keywords='gearthonic',
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: Implementation :: PyPy',
    ],
)



"""Tests for the protocols."""
from _ssl import CERT_REQUIRED, CERT_NONE

import pytest
from mock import Mock, patch

from gearthonic.exceptions import ConfigurationError
from gearthonic.protocols import XmlRpcProtocol, JsonRpcProtocol


class TestXmlProtocol(object):
    """Tests for the XML RPC protocol."""

    @patch('gearthonic.protocols.ServerProxy')
    def test_server_proxy_configuration(self, server_mock):
        XmlRpcProtocol('example.com', 1234)
        server_mock.assert_called()
        args, kwargs = server_mock.call_args_list[0]
        assert args[0] == 'https://example.com:1234'
        assert kwargs['context'].verify_mode == CERT_REQUIRED

    @patch('gearthonic.protocols.ServerProxy')
    def test_insecure_server_proxy_configuration(self, server_mock):
        XmlRpcProtocol('example.com', 1234, secure=False)
        server_mock.assert_called_once_with('http://example.com:1234', context=None)
        server_mock.reset_mock()

        XmlRpcProtocol('example.com', 1234, verify=False)
        server_mock.assert_called()
        args, kwargs = server_mock.call_args_list[0]
        assert args[0] == 'https://example.com:1234'
        assert kwargs['context'].verify_mode == CERT_NONE

    @patch('gearthonic.protocols.ServerProxy')
    def test_authentication(self, server_mock):
        XmlRpcProtocol('example.com', 1234, username='ham', password='eggs')
        server_mock.assert_called()
        args, kwargs = server_mock.call_args_list[0]
        assert args[0] == 'https://ham:eggs@example.com:1234'

    def test_authentication_with_insecure_connection(self):
        with pytest.raises(ConfigurationError):
            XmlRpcProtocol('example.com', 1234, secure=False, username='ham', password='eggs')

    @patch('gearthonic.protocols.ServerProxy', new=Mock())
    def test_call_method(self):
        xrp = XmlRpcProtocol('example.com', 1234)
        xrp.call('listMethods', 'foobar', ham='eggs')
        xrp.proxy.listMethods.assert_called()
        xrp.proxy.listMethods.assert_called_once_with('foobar', ham='eggs')


class TestJsonProtocol(object):
    """Tests for the JSON RPC protocol."""

    def test_http_connection_configuration(self):
        jrp = JsonRpcProtocol('example.com', 1234)
        assert jrp.uri == 'https://example.com:1234'
        assert jrp.server.endpoint == 'https://example.com:1234'
        assert jrp.server.session.verify
        assert not jrp.authentication

    def test_insecure_http_connection_configuration(self):
        jrp = JsonRpcProtocol('example.com', 1234, secure=False)
        assert jrp.uri == 'http://example.com:1234'
        assert jrp.server.endpoint == 'http://example.com:1234'
        assert jrp.server.session.verify
        assert not jrp.authentication

        jrp = JsonRpcProtocol('example.com', 1234, verify=False)
        assert jrp.uri == 'https://example.com:1234'
        assert jrp.server.endpoint == 'https://example.com:1234'
        assert not jrp.server.session.verify
        assert not jrp.authentication

    def test_authentication_configuration(self):
        jrp = JsonRpcProtocol('example.com', 1234, username='ham', password='eggs')
        assert jrp.uri == 'https://example.com:1234'
        assert jrp.server.endpoint == 'https://example.com:1234'
        assert jrp.server.session.verify
        assert jrp.authentication
        assert jrp.server.session.auth == ('ham', 'eggs')

    def test_authentication_with_insecure_connection(self):
        with pytest.raises(ConfigurationError):
            JsonRpcProtocol('example.com', 1234, secure=False, username='ham', password='eggs')



"""Tests for the collections"""

from mock import Mock

from gearthonic.methods import BaseCollection


class TestBaseCollection(object):
    """Tests the BaseCollection"""

    def test_collection_uses_call_method(self, gear_client, monkeypatch):
        """Tests that the BaseCollection uses the call method of the GearClient"""
        call_mock = Mock()
        # Mock the call method of the GearClient to check, if it got called
        monkeypatch.setattr(gear_client, 'call', call_mock)
        bc = BaseCollection(gear_client)
        bc.call('list_all', 1, ham='eggs')

        assert call_mock.called
        call_mock.assert_called_once_with('list_all', 1, ham='eggs')
